import React, { ReactNode, forwardRef, useCallback } from 'react';
import {
  Platform,
  StyleProp,
  Text,
  TextStyle,
  requireNativeComponent,
} from 'react-native';
import { v4 } from 'uuid';
import {
  Highlight,
  HighlightRectsCalculatedNativeEvent,
  RNSelectableTextProps,
  Selection,
  SelectionChangeNativeEvent,
  onWordPressIOSNativeEvent,
} from './types';
import { combineHighlights, mapHighlightsRanges } from './utils';

const RNSelectableText =
  requireNativeComponent<RNSelectableTextProps>('RNSelectableText');

interface HighlightableTextProps {
  value: string;
  style?: StyleProp<TextStyle>;
  highlights: Highlight[];
  highlightStyle?: StyleProp<TextStyle>;
  onDismissSelection?: () => void;
  onSelectionChange?: (selection: Selection) => void;
  onHighlightPress?: (id: string) => void;
  onHighlightRectsCalculated?: (
    event: HighlightRectsCalculatedNativeEvent
  ) => void;
}

export type HighlightableTextRef = {};

export const HighlightableText = forwardRef<
  HighlightableTextRef,
  HighlightableTextProps
>((props, ref) => {
  const {
    value,
    style,
    highlights,
    highlightStyle,
    onDismissSelection,
    onSelectionChange,
    onHighlightPress,
    onHighlightRectsCalculated,
  } = props;

  const onTextSelectionChange = useCallback(
    (event: SelectionChangeNativeEvent) => {
      onSelectionChange?.(event.nativeEvent.selection);
    },
    [onSelectionChange]
  );

  const onWordPressNative = useCallback(
    (event: onWordPressIOSNativeEvent) => {
      if (highlights.length === 0) {
        onDismissSelection?.();
        return;
      }

      const { clickedRangeStart, clickedRangeEnd } = event.nativeEvent;
      const mergedHighlights = combineHighlights(highlights);

      const highlightInRange = mergedHighlights.find(
        ({ start, end }) =>
          clickedRangeStart >= start - 1 && clickedRangeEnd <= end + 1
      );

      if (highlightInRange) {
        onHighlightPress?.(highlightInRange.id);
      } else {
        onDismissSelection?.();
      }
    },
    [highlights, onDismissSelection, onHighlightPress]
  );

  let textValue: ReactNode[] = [value];

  if (highlights.length > 0) {
    textValue = mapHighlightsRanges(value, highlights).map(
      ({ id, isHighlight, text }) => (
        <Text
          key={v4()}
          selectable
          style={isHighlight ? highlightStyle : {}}
          onPress={isHighlight ? () => onHighlightPress?.(id ?? '') : undefined}
        >
          {text}
        </Text>
      )
    );
  }

  return (
    <RNSelectableText
      ref={ref}
      // change key to force re-render whenever highlights change
      key={JSON.stringify(highlights)}
      style={style}
      highlights={highlights}
      onWordPress={Platform.OS === 'ios' ? onWordPressNative : () => {}}
      onTextSelectionChange={onTextSelectionChange}
      onHighlightRectsCalculated={onHighlightRectsCalculated}
    >
      <Text>{textValue}</Text>
    </RNSelectableText>
  );
});
